package com.dobsondev;

import java.io.FileNotFoundException;
import java.util.ArrayList;

/**
 * This class is used to as the main class in the A* program. It handles
 * the input from the user via command line, and then performs the actual
 * A* algorithm on that input. The input is made into objects Map and Node,
 * whose classes are separate from this one.
 * 
 * @author Alex Dobson
 *         University of Alberta, Augustana Campus
 *         2013-04-01
 *
 */
public class AStar {
	
	private static Map map;
	private static int start_x, start_y, goal_x, goal_y;
	private static boolean print_map;
	
	/**
	 * The main method handles the input handling logic as well as calls to the
	 * other methods in order to properly perform the A* pathfinding algorithm.
	 * 
	 * @param args - arguments passed in through the command line
	 */
	public static void main(String[] args) {
		
		switch(args.length) {
		
		//Invalid number of arguments or looking for usage message.
		case 1:
			if (args[0].equals("--h")) {
				System.out.println("USAGE: java Main map_file start_node goal_node");
				System.out.println("   eg. java Main res/map_1.txt (1,1) (1,7)");
				System.out.println("   eg. java Main res/map_1.txt (1,1) (1,7) --no_map");
				System.exit(0);
			} else {
				System.out.println("INVALID ARGUEMENTS");
				System.out.println("USAGE: java Main map_file start_node goal_node");
				System.out.println("   eg. java Main res/map_1.txt (1,1) (1,7)");
				System.out.println("   eg. java Main res/map_1.txt (1,1) (1,7) --no_map");
				System.exit(0);
			}
			break;
			
		// Valid number of arguments and print map	
		case 3:
			tryMap(args[0]);
			tryStartNode(args[1]);
			tryGoalNode(args[2]);
			print_map = true;
			break;
		
		// Valid number of arguments and do not print the map
		case 4:
			tryMap(args[0]);
			tryStartNode(args[1]);
			tryGoalNode(args[2]);
			if (args[3].equals("--no_map")) {
				print_map = false;
			} else {
				System.out.println("INVALID ARGUEMENTS");
				System.out.println("USAGE: java Main map_file start_node goal_node");
				System.out.println("   eg. java Main res/map_1.txt (1,1) (1,7)");
				System.out.println("   eg. java Main res/map_1.txt (1,1) (1,7) --no_map");
				System.exit(0);
			}
			break;
		//Invalid number of arguments	
		default:
			System.out.println("INVALID ARGUEMENTS");
			System.out.println("USAGE: java Main map_file start_node goal_node");
			System.out.println("   eg. java Main res/map_1.txt (1,1) (1,7)");
			System.out.println("   eg. java Main res/map_1.txt (1,1) (1,7) --no_map");
			System.exit(0);
			break;
			
		}//end switch(args.length)
		
		System.out.println("+");
		System.out.println("| Map File: '" + map.getFileName() + "'");
		System.out.println("| Start Node: (" + start_x + ", " + start_y + ")");
		System.out.println("| Goal Node: (" + goal_x + ", " + goal_y + ")");
		System.out.println("+");
		
		for (int run = 0; run < 3; run++) {
			a_star_pathfinder(run);
		}
		
		System.out.println("Exiting...");
	}//end main
	
	/**
	 * This method is used to set the start_x and start_y values for the entire
	 * program. The method is given a command line argument string as input, and
	 * then parses the string to get the numeric values from it. If these values
	 * are valid, then the method sets the start_x and start_y values, otherwise
	 * it terminates the program.
	 * 
	 * @param start - the start node command line argument
	 */
	public static void tryStartNode(String start) {
		String delims = "[( , )]";
		String[] start_node = start.split(delims);
		
		int temp_start_x = Integer.parseInt(start_node[1]);
		int temp_start_y = Integer.parseInt(start_node[2]);
		
		if (map.getNode(temp_start_x, temp_start_y).isTraversable() 
				&& temp_start_x < map.getWidth()
				&& temp_start_x >= 0
				&& temp_start_y < map.getHeight()
				&& temp_start_y >= 0) {
			start_x = temp_start_x;
			start_y = temp_start_y;
		} else {
			System.out.println("INVALID START NODE: " + start);
			System.exit(0);
		}
	}
	
	/**
	 * This method is used to set the goal_x and goal_y values for the entire
	 * program. The method is given a command line argument string as input, and
	 * then parses the string to get the numerical values from it. If these values
	 * are valid, then the method sets the goal_x and goal_y values, otherwise it
	 * terminates the program.
	 * 
	 * @param goal - the goal node command line argument
	 */
	public static void tryGoalNode(String goal) {
		String delims = "[( , )]";
		String[] goal_node = goal.split(delims);
		
		int temp_goal_x = Integer.parseInt(goal_node[1]);
		int temp_goal_y = Integer.parseInt(goal_node[2]);
		
		if (map.getNode(temp_goal_x, temp_goal_y).isTraversable()
				&& temp_goal_x < map.getWidth()
				&& temp_goal_x >= 0
				&& temp_goal_y < map.getHeight()
				&& temp_goal_y >= 0) {
			goal_x = temp_goal_x;
			goal_y = temp_goal_y;
		} else {
			System.out.println("INVALID GOAL NODE: " + goal);
			System.exit(0);
		}
	}
	
	/**
	 * This method attempts to open the map file passed in as a command line argument
	 * to the program. If the map file exists, then the method sets the value of map
	 * to a new Map object with the file as a parameter. If the file does not exist,
	 * then the program terminates.
	 * 
	 * @param map_file - the map file command line argument
	 */
	public static void tryMap(String map_file) {
		try {
			map = new Map(map_file);
		} catch (FileNotFoundException e) {
			System.out.println("INVALID MAP FILE " + map_file);
			System.exit(0);
		}
	}//end tryMap
	
	/**
	 * This is the real main method of this program. This method performs the A*
	 * path finding algorithm. It takes as input only an integer representing what
	 * heuristic it is supposed to use. The algorithm uses a best-first search and
	 * then finds a least-cost path from the start node to the goal node. As the
	 * algorithm runs, it follows the path which it expects to give the lowest total
	 * cost to the goal. It uses a knowledge-plus-heuristic cost function to calculate
	 * this path. In the algorithm, the h-value is the heuristic (or expected cost)
	 * from the node to the goal node. By adding this to the g-value for each node,
	 * which is the cost to travel to that node, we can find the shortest path.
	 * 
	 * @param heuristic_option - the heuristic to be used by the algorithm
	 */
	public static void a_star_pathfinder(int heuristic_option) {
		long start_time =  System.currentTimeMillis();
		
		// Run the algorithm with different heuristics
		switch(heuristic_option) {
		case 0:
			setManhanttanDistances();
			break;
		case 1:
			setEuclideanDistances();
			break;
		case 2:
			setDiagonalDistances();
			break;
		default:
			setManhanttanDistances();
			break;
		}//end switch(heuristic_option)
		
		
		// The closed list is the set of nodes already checked
		ArrayList<Node> closed_list = new ArrayList<Node>();
		
		// The open list is the set of nodes to be checked, at first
		// it only contains the start node.
		ArrayList<Node> open_list = new ArrayList<Node>();
		open_list.add(map.getNode(start_x, start_y));
		
		map.getNode(start_x, start_y).setG_value(0);
		map.getNode(start_x, start_y).setF_value();
		
		boolean answer_found = false;
		
		while(!open_list.isEmpty() && !answer_found) {
			Node current = getNewCurrentNode(open_list);
			open_list.remove(current);
			closed_list.add(current);
			
			// We found the goal, so we reconstruct the path by following the parent
			// starting at the goal node.
			if (current.getX() == goal_x && current.getY() == goal_y) {
				long end_time = System.currentTimeMillis();
				markPath();
				// Output with map
				if (print_map) {
					System.out.println("The Shortest Path using " + getHeuristicName(heuristic_option) + " is: ");
					map.print();
					System.out.println("Time: " + (end_time - start_time) + "ms");
					System.out.println("");
				// Output with no map
				} else {
					System.out.println("Time using " + getHeuristicName(heuristic_option) + " is: " 
								+ (end_time - start_time) + "ms");
					System.out.println("");
				}
				unmarkPath();
				answer_found = true;
			}
			
			if (!answer_found) {
				// Check the current nodes neighbors to see if there is a smaller
				// g-value including the distance to travel from the current node
				// to the neighbor node. If there is, then we change that node's
				// parent to the current node since the shortest path to that
				// node is actually by going through the current node.
				Node[] neighbors = getNeighbors(current);
				for (int i = 0; i < neighbors.length; i++) {
					// The neighbor has to be traversable to be considered
					if (neighbors[i].isTraversable()) {
						double temp_g_value = neighbors[i].getG_value() + 1;
						if (closed_list.contains(neighbors[i])) {
							if (temp_g_value >= neighbors[i].getG_value()) {
								continue;
							}
						}
						// If the neighbor is not in the open set then we set its parent to our
						// current node, and set its g-value and f-value accordingly. If the
						// neighbor is in the open set, but its current g-value is greater than
						// temp g-value (this means its faster to get to the neighbor by going
						// through our current node) then we set the neighbors parent to our
						// current node and set the g-value and f-value accordingly.
						if (!open_list.contains(neighbors[i]) || temp_g_value < neighbors[i].getG_value()) {
							neighbors[i].setParent(current);
							neighbors[i].setG_value(temp_g_value);
							neighbors[i].setF_value();
							if (!open_list.contains(neighbors[i])) {
								open_list.add(neighbors[i]);
							}
						}
					}
				}
			}
		}
		// Depending on the map file, there may not exist a path from the start to the end
		// node for A* to find. In this case we simply print that we could not find an
		// answer, most likely because there is no path.
		if (!answer_found) {
			System.out.println("A* Could not find an answer.");
			System.out.println("There is most likely no path from " + "(" + start_x + ", " 
					+ start_y + "to (" + goal_x + ", " + goal_y + ")");
		}
	}//end AStar
	
	/**
	 * This method finds neighbor nodes of the current node being examined by A*. It
	 * then returns all the neighbors in an array for easy access by the main algorithm.
	 * 
	 * @param node - the current node
	 * @return neighbors[] - the array of node's neighbors
	 */
	public static Node[] getNeighbors(Node node) {
		Node[] neighbors;
		// Top left corner
		if (node.getX() == 0 && node.getY() == 0) {
			neighbors = new Node[2];
			neighbors[0] = map.getNode(node.getX() + 1, node.getY());
			neighbors[1] = map.getNode(node.getX(), node.getY() + 1);
		} 
		// Top right corner
		else if (node.getX() == 0 && node.getY() == (map.getHeight() - 1)) {
			neighbors = new Node[2];
			neighbors[0] = map.getNode(node.getX() + 1, node.getY());
			neighbors[1] = map.getNode(node.getX(), node.getY() - 1);
		}
		//bottom left corner
		else if (node.getX() == (map.getWidth() - 1) && node.getY() == 0) {
			neighbors = new Node[2];
			neighbors[0] = map.getNode(node.getX() - 1, node.getY());
			neighbors[1] = map.getNode(node.getX(), node.getY() + 1);
		}
		// Bottom right corner
		else if (node.getX() == (map.getWidth() - 1) && node.getY() == (map.getHeight() - 1)) {
			neighbors = new Node[2];
			neighbors[0] = map.getNode(node.getX() - 1, node.getY());
			neighbors[1] = map.getNode(node.getX(), node.getY() - 1);
		}
		// Bottom row, but not a corner
		else if (node.getX() == 0) {
			neighbors = new Node[3];
			neighbors[0] = map.getNode(node.getX() + 1, node.getY());
			neighbors[1] = map.getNode(node.getX(), node.getY() + 1);
			neighbors[2] = map.getNode(node.getX(), node.getY() - 1);
		}
		// Top row, but not a corner
		else if (node.getX() == (map.getWidth() - 1)) {
			neighbors = new Node[3];
			neighbors[0] = map.getNode(node.getX() - 1, node.getY());
			neighbors[1] = map.getNode(node.getX(), node.getY() + 1);
			neighbors[2] = map.getNode(node.getX(), node.getY() - 1);
		}
		// Left column, but not a corner
		else if (node.getY() == 0) {
			neighbors = new Node[3];
			neighbors[0] = map.getNode(node.getX(), node.getY() + 1);
			neighbors[1] = map.getNode(node.getX() + 1, node.getY());
			neighbors[2] = map.getNode(node.getX() - 1, node.getY());
		}
		// Right column, but not a corner
		else if (node.getY() == (map.getHeight() - 1)) {
			neighbors = new Node[3];
			neighbors[0] = map.getNode(node.getX(), node.getY() - 1);
			neighbors[1] = map.getNode(node.getX() + 1, node.getY());
			neighbors[2] = map.getNode(node.getX() - 1, node.getY());
		}
		// All other nodes not on the sides
		else {
			neighbors = new Node[4];
			neighbors[0] = map.getNode(node.getX() + 1, node.getY());
			neighbors[1] = map.getNode(node.getX() - 1, node.getY());
			neighbors[2] = map.getNode(node.getX(), node.getY() + 1);
			neighbors[3] = map.getNode(node.getX(), node.getY() - 1);
		}
		return neighbors;
	}//end getNeighbors
	
	/**
	 * This method returns the new node to be examined by A*. A* always checks
	 * the node with the lowest f-value in the open list, so this method simply
	 * returns that node.
	 * 
	 * @param open_list - the open list from A*
	 * @return current - the new current node to be examined by A*
	 */
	public static Node getNewCurrentNode(ArrayList<Node> open_list) {
		Node current = open_list.get(0);
		for (int i = 1; i < open_list.size(); i++) {
			if (current.getF_value() > open_list.get(i).getF_value()) {
				current = open_list.get(i);
			}
		}
		return current;
	}//end getNewCurrentNode
	
	/**
	 * This method marks the path from the goal node when we have found it back
	 * to the start node so we know the best path. The method achieves this by
	 * tracing back the parent nodes from the goal node.
	 */
	public static void markPath() {
		boolean done = false;
		Node current = map.getNode(goal_x, goal_y);
		while(!done) {
			map.getNode(current.getX(), current.getY()).setInPath(true);
			if (current.getX() == start_x && current.getY() == start_y) {
				done = true;
			} else {
				current = current.getParent();
			}
		}
	}//end markPath
	
	/**
	 * This method unmarks the path from the goal node to the start node so
	 * that on the next run of the algorithm, if A* takes a different path,
	 * we won't have more than one path being printed at the same time.
	 */
	public static void unmarkPath() {
		for (int x = 0; x < map.getWidth(); x++) {
			for (int y = 0; y < map.getHeight(); y++) {
				map.getNode(x, y).setInPath(false);
			}
		}
	}
	
/******************************* HEURISTICS *******************************/
	public static void setManhanttanDistances() {
		for (int x = 0; x < map.getWidth(); x++) {
			for (int y = 0; y < map.getHeight(); y++) {
				double dx = Math.abs(x - goal_x);
				double dy = Math.abs(y - goal_y);
				map.getNode(x, y).setH_value(dx + dy);
			}
		}
	}//end setManhattanDistance
	
	public static void setEuclideanDistances() {
		for (int x = 0; x < map.getWidth(); x++) {
			for (int y = 0; y < map.getHeight(); y++) {
				int dx = (x - goal_x) ^ 2;
				int dy = (y - goal_y) ^ 2;
				map.getNode(x, y).setH_value(Math.sqrt(dx + dy));
			}
		}
	}//end setEuclideanDistances
	
	public static void setDiagonalDistances() {
		for (int x = 0; x < map.getWidth(); x++) {
			for (int y = 0; y < map.getHeight(); y++) {
				double dx = Math.abs(x - goal_x);
				double dy = Math.abs(y - goal_y);
				map.getNode(x, y).setH_value(Math.max(dx, dy));
			}
		}
	}//end setDiagonalDistances
	
	public static String getHeuristicName(int heuristic_option) {
		switch(heuristic_option) {
		case 0:
			return "Manhattan Distance";
		case 1:
			return "Euclidean Distance";
		case 2:
			return "Diagonal Distance";
		default:
			return "INVALID HEURISTIC";
		}//end switch(heuristic_option)
	}
/***************************************************************************/
}//end public class Main